import rdflib
import re


class RdfGraph(object):
    """Summary

    Attributes
    ----------
    gprefix : rdflib.namespace.Namespace
        Rdf prefix for askomics
    graph : rdflib.Graph
        The RDF graph
    """

    def __init__(self, askomics_prefix):
        """init

        Parameters
        ----------
        askomics_prefix : str
            AskOmics prefix
        """
        self.gprefix = rdflib.namespace.Namespace(askomics_prefix)
        self.graph = rdflib.Graph()

        self.graph.bind('askomics', askomics_prefix)

    def check_entity(self, entity):
        """Check if entity is correct (not rdf rdfs owl or virtuoso thing)

        Parameters
        ----------
        entity : str
            Entity uri to check

        Returns
        -------
        bool
            True if entity is a true one
        """
        excluded_prefixes = (
            "http://www.w3.org/2002/07/owl#",
            "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
            "http://www.w3.org/2000/01/rdf-schema#",
            "http://www.openlinksw.com/schemas/virtrdf#",
            "http://www.w3.org/ns/sparql-service-description#"
        )

        if entity.lower().startswith(excluded_prefixes):
            return False
        return True

    def add_entities_and_relations(self, sparql_result):
        """Add entities and relation in the rdf graph

        Parameters
        ----------
        sparql_result : list
            Sparql result
        """
        entities = []

        # Entities and relations
        for result in sparql_result:
            source_entity = result["source_entity"]
            target_entity = result["target_entity"]
            relation = result["relation"]

            # Source entity
            if self.check_entity(source_entity) and source_entity not in entities:
                entities.append(source_entity)
                self.graph.add((rdflib.URIRef(source_entity), rdflib.RDF.type, self.gprefix["entity"]))
                self.graph.add((rdflib.URIRef(source_entity), rdflib.RDF.type, self.gprefix["startPoint"]))
                self.graph.add((rdflib.URIRef(source_entity), rdflib.RDF.type, rdflib.OWL.Class))
                self.graph.add((rdflib.URIRef(source_entity), self.gprefix["instancesHaveNoLabels"], rdflib.Literal(True)))
                self.graph.add((rdflib.URIRef(source_entity), rdflib.RDFS.label, rdflib.Literal(self.get_label(source_entity))))

            # Target entity
            if self.check_entity(target_entity) and target_entity not in entities:
                entities.append(target_entity)
                self.graph.add((rdflib.URIRef(target_entity), rdflib.RDF.type, self.gprefix["entity"]))
                self.graph.add((rdflib.URIRef(target_entity), rdflib.RDF.type, self.gprefix["startPoint"]))
                self.graph.add((rdflib.URIRef(target_entity), rdflib.RDF.type, rdflib.OWL.Class))
                self.graph.add((rdflib.URIRef(target_entity), self.gprefix["instancesHaveNoLabels"], rdflib.Literal(True)))
                self.graph.add((rdflib.URIRef(target_entity), rdflib.RDFS.label, rdflib.Literal(self.get_label(target_entity))))

            # Relation
            if self.check_entity(relation):
                self.graph.add((rdflib.URIRef(relation), rdflib.RDF.type, rdflib.OWL.ObjectProperty))
                self.graph.add((rdflib.URIRef(relation), rdflib.RDF.type, self.gprefix["AskomicsRelation"]))
                self.graph.add((rdflib.URIRef(relation), rdflib.RDFS.label, rdflib.Literal(self.get_label(relation))))
                self.graph.add((rdflib.URIRef(relation), rdflib.RDFS.domain, rdflib.URIRef(source_entity)))
                self.graph.add((rdflib.URIRef(relation), rdflib.RDFS.range, rdflib.URIRef(target_entity)))

    def add_decimal_attributes(self, sparql_result):
        """Add decimal  in the rdf graph

        Parameters
        ----------
        sparql_result : list
            Sparql result
        """
        for result in sparql_result:
            entity = result["entity"]
            attribute = result["attribute"]

            if self.check_entity(entity) and self.check_entity(attribute):
                """<FRESHLY_INSERTED>"""
                self.graph.add((rdflib.URIRef(attribute), rdflib.RDF.type, rdflib.OWL.DatatypeProperty))
                self.graph.add((rdflib.URIRef(attribute), rdflib.RDFS.label, rdflib.Literal(self.get_label(attribute))))
                self.graph.add((rdflib.URIRef(attribute), rdflib.RDFS.domain, rdflib.URIRef(entity)))
                self.graph.add((rdflib.URIRef(attribute), rdflib.RDFS.range, rdflib.XSD.decimal))

    def add_text_attributes(self, sparql_result):
        """Add text  in the rdf graph

        Parameters
        ----------
        sparql_result : list
            Sparql result
        """
        for result in sparql_result:
            entity = result["entity"]
            attribute = result["attribute"]

            if self.check_entity(entity) and self.check_entity(attribute):

                self.graph.add((rdflib.URIRef(attribute), rdflib.RDF.type, rdflib.OWL.DatatypeProperty))
                self.graph.add((rdflib.URIRef(attribute), rdflib.RDFS.label, rdflib.Literal(self.get_label(attribute))))
                self.graph.add((rdflib.URIRef(attribute), rdflib.RDFS.domain, rdflib.URIRef(entity)))
                self.graph.add((rdflib.URIRef(attribute), rdflib.RDFS.range, rdflib.XSD.string))

    def get_label(self, uri):
        """Get a label from an URI

        Parameters
        ----------
        uri : string
            URi to get a label

        Returns
        -------
        string
            Label
        """
        return self.uncamel(uri.split("/")[-1].split("#")[-1].replace("_", " "))

    @staticmethod
    def uncamel(string):
        """Insert space beween camelcased words

        Parameters
        ----------
        string : str
            CamelCased string

        Returns
        -------
        str
            Uncamel Cased string
        """
        re_outer = re.compile(r'([^A-Z ])([A-Z])')
        re_inner = re.compile(r'\b[A-Z]+(?=[A-Z][a-z])')
        uncameled = re_inner.sub(r'\g<0> ', re_outer.sub(r'\1 \2', string))
        if uncameled[0].islower():
            return uncameled.lower()
        return uncameled



import textwrap


class QueryLibrary(object):
    """SPARQL methods"""

    def __init__(self):
        """init"""
        pass

    @property
    def entities_and_relations(self):
        """Sparql query to get entities and relations

        Returns
        -------
        str
            SPARQL query
        """
        return textwrap.dedent('''
        SELECT DISTINCT ?source_entity ?relation ?target_entity
        WHERE {
            # Get entities
            ?instance_of_source a ?source_entity .
            ?instance_of_target a ?target_entity .
            # Relations
            ?instance_of_source ?relation ?instance_of_target .
        }
        ''')

    @property
    def entities_and_numeric_attributes(self):
        """Sparql query to get entities and numeric attributes

        Returns
        -------
        str
            SPARQL query
        """
        return textwrap.dedent('''
        SELECT DISTINCT ?entity ?attribute
        WHERE {
            # Get entities
            ?instance_of_entity a ?entity .
            # Attributes
            ?instance_of_entity ?attribute ?value .
            FILTER (isNumeric(?value))
        }
        ''')

    @property
    def entities_and_text_attributes(self):
        """Sparql query to get entities and text attributes

        Returns
        -------
        str
            SPARQL query
        """
        return textwrap.dedent('''
        SELECT DISTINCT ?entity ?attribute
        WHERE {
            # Get entities
            ?instance_of_entity a ?entity .
            # Attributes
            ?instance_of_entity ?attribute ?value .
            FILTER (isLiteral(?value))
            FILTER (!isNumeric(?value))
        }
        ''')

    @property
    def ontologies(self):
        """Sparql query to get ontologies

        Returns
        -------
        str
            SPARQL query
        """
        return textwrap.dedent('''
        SELECT DISTINCT ?ontology
        WHERE {
            ?ontology a owl:Ontology .
            ?entity a owl:Class .
            ?relation a owl:ObjectProperty .
            ?attribute a owl:DatatypeProperty .
            ?entity rdfs:isDefinedBy ?ontology .
            ?relation rdfs:isDefinedBy ?ontology .
            ?attribute rdfs:isDefinedBy ?ontology .
        }
        ''')

    @staticmethod
    def entities_and_relations_with_ontology(ontology):
        """Sparql query to get entities and relations

        Returns
        -------
        str
            SPARQL query
        """
        return textwrap.dedent('''
        SELECT DISTINCT ?source_entity ?relation ?target_entity
        WHERE {{
            ?source_entity a owl:Class .
            ?source_entity rdfs:isDefinedBy <{ontology}> .

            ?target_entity a owl:Class .
            ?target_entity rdfs:isDefinedBy <{ontology}> .

            ?relation a owl:ObjectProperty .
            ?relation rdfs:domain ?source_entity .
            ?relation rdfs:range ?target_entity .
        }}
        '''.format(ontology=ontology))

    @staticmethod
    def entities_and_numeric_attributes_with_ontology(ontology):
        """Sparql query to get entities and numeric attributes

        Returns
        -------
        str
            SPARQL query
        """
        return textwrap.dedent('''
        SELECT DISTINCT ?entity ?attribute
        WHERE {{
            # Entity
            ?entity a owl:Class .
            ?entity rdfs:isDefinedBy <{ontology}> .
            # Attribute
            ?attribute a owl:DatatypeProperty .
            ?attribute rdfs:domain ?entity .
            ?attribute rdfs:range ?range .
            VALUES ?range {{ xsd:float xsd:int }} .
        }}
        '''.format(ontology=ontology))

    @staticmethod
    def entities_and_text_attributes_with_ontology(ontology):
        """Sparql query to get entities and numeric attributes

        Returns
        -------
        str
            SPARQL query
        """
        return textwrap.dedent('''
        SELECT DISTINCT ?entity ?attribute
        WHERE {{
            # Entity
            ?entity a owl:Class .
            ?entity rdfs:isDefinedBy <{ontology}> .
            # Attribute
            ?attribute a owl:DatatypeProperty .
            ?attribute rdfs:domain ?entity .
            ?attribute rdfs:range ?range .
            VALUES ?range {{ xsd:string }} .
        }}
        '''.format(ontology=ontology))



import rdflib
from SPARQLWrapper import SPARQLWrapper, JSON


class SparqlQuery(object):
    """SPARQL methods

    Attributes
    ----------
    prefix : TYPE
        Description
    prefixes : TYPE
        Description
    rdf_source : TYPE
        Description
    source : TYPE
        Description
    source_type : TYPE
        Description
    """

    def __init__(self, source, source_type, prefix):
        """Init

        Parameters
        ----------
        source : TYPE
            Description
        source_type : TYPE
            Description
        prefix : string
            Prefix URI
        """
        self.source = source
        self.source_type = source_type
        self.prefix = prefix
        self.prefixes = {
            "owl:": "http://www.w3.org/2002/07/owl#",
            "rdf:": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
            "rdfs:": "http://www.w3.org/2000/01/rdf-schema#",
            "dc:": "http://purl.org/dc/elements/1.1/",
            "prov:": "http://www.w3.org/ns/prov#",
            "xsd:": "http://www.w3.org/2001/XMLSchema#",
            "skos:": "http://www.w3.org/2004/02/skos/core#",
            "chebi:": "http://purl.obolibrary.org/obo/",
            "drugbankdrugs:": "http://wifo5-04.informatik.uni-mannheim.de/drugbank/resource/drugs/",
            "askomics:": self.prefix
        }

        # if source is a file, load it in a rdflib graph
        self.rdf_source = None
        if self.source_type != "sparql":
            self.rdf_source = rdflib.Graph()
            self.rdf_source.parse(self.source, format=self.source_type)

    def get_sparl_prefix(self):
        """Get a SPARQL prefix string

        Returns
        -------
        str
            SPARQL prefix string
        """
        prefixes_string = ""
        for prefix, uri in self.prefixes.items():
            prefixes_string += "PREFIX {} <{}>\n".format(prefix, uri)

        return prefixes_string

    def get_ttl_prefix(self):
        """Get a turtle prefix string

        Returns
        -------
        str
            Turtle prefix string
        """
        prefixes_string = ""
        for prefix, uri in self.prefixes.items():
            prefixes_string += "@prefix {} <{}> .\n".format(prefix, uri)

        return prefixes_string

    def execute_sparql_query(self, query):
        """Execute query on a SPARQL endpoint

        Parameters
        ----------
        query : str
            The query

        Returns
        -------
        list
            results
        """
        endpoint = SPARQLWrapper(self.source)
        endpoint.setQuery(query)
        endpoint.setReturnFormat(JSON)
        return endpoint.query().convert()

    def execute_rdflib_query(self, query):
        """Execute query on a rdflib graph

        Parameters
        ----------
        query : str
            The query

        Returns
        -------
        list
            results
        """
        return self.rdf_source.query(query)

    def parse_sparql_results(self, json_results):
        """Parse result of sparql query

        Parameters
        ----------
        json_results : dict
            Query result

        Returns
        -------
        list
            Parsed results
        """
        try:
            data = []
            for row in json_results["results"]["bindings"]:
                row_dict = {}
                for key, value in row.items():
                    row_dict[key] = value['value']
                data.append(row_dict)

        except Exception as e:
            print(str(e))
            return []

        return data

    def parse_rdflib_results(self, results):
        """Parse result of sparql query (rdflib)

        Parameters
        ----------
        json_results : dict
            Query result

        Returns
        -------
        list
            Parsed results
        """
        variables = [str(v) for v in results.vars]
        data = []
        for row in results:
            row_dict = {}
            for v in variables:
                row_dict[v] = str(row[v])
            data.append(row_dict)

        return data

    def process_query(self, query):
        """Execute a query and return parsed results

        Parameters
        ----------
        query : string
            The query to execute

        Returns
        -------
        list
            Parsed results
        """
        # prefixed_query = self.get_sparl_prefix() + query
        # print(query)
        if self.source_type == "sparql":
            return self.parse_sparql_results(self.execute_sparql_query(query))
        else:
            return self.parse_rdflib_results(self.execute_rdflib_query(query))



from setuptools import setup, find_packages

setup(
    name='abstractor',
    version='2.0.0',
    description='Abstraction generator for AskOmics, from a distant SPARQL endpoint',
    author='Xavier Garnier',
    author_email='xavier.garnier@irisa.fr',
    url='https://github.com/askomics/abstractor',
    download_url='https://github.com/askomics/abstractor/archive/2.0.0.tar.gz',
    install_requires=['SPARQLWrapper'],
    packages=find_packages(),
    license='AGPL',
    platforms='Posix; MacOS X; Windows',
    scripts=['abstractor'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.0',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ])
